
# THIS FILE IS GENERATED FROM ORANGE SETUP.PY
short_version = '3.38.5'
version = '3.38.5'
full_version = '3.38.5.dev0+2fbbf2c'
git_revision = '2fbbf2cc8436e6799de543bd2758774776acb4e5'
release = False

if not release:
    version = full_version
