"""
Widgets from Evaluate category

"""
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.evaluate.evaluate." + key)



NAME = "Evaluate"
ID = "orange.widgets.evaluate"
LABEL = __("package_label")

DESCRIPTION = __("package_desc")

BACKGROUND = "#C3F3F3"

ICON = "icons/Category-Evaluate.svg"

PRIORITY = 5
